from typing import List, Tuple

import numpy as np
import statistics

import pandas as pd
from scipy import signal
from scipy.fftpack import fft
from scipy.stats import skew

from .core.EEG_freqbands import eeg_freq_bands


def mean_val(signal_list: List[int or float]) -> float or int:
	"""
	Function that find the mean of a give array

	:param signal_list: a 1d array of edf signals
	:return: The mean of the array
	"""
	length = len(signal_list)
	if length == 0:
		return 0.0

	return float(sum(signal_list) / length)


def std_val(signal_list: List[int or float]) -> float:
	"""
	Function that finds the given standard deviation of a given array

	:param signal_list:  a 1d array of edf signals
	:return: The standard deviation of the array
	"""
	if len(signal_list) < 2:
		raise Exception("Your array must contain at least two values")
	return statistics.stdev(signal_list)


def fourier(epochized_data: List[int or float], sampling_freq: int or float) -> List[np.ndarray]:
	"""

	Args:
		epochized_data:
		sampling_freq:

	Returns:

	"""
	return_list = []
	T = 1 / sampling_freq
	for i in epochized_data:
		yf = fft(i)
		yf = np.array(yf)
		return_list.append(yf)
	return return_list


def welch_psd(epoch: pd.DataFrame, fs: int or float, filter_order: int)\
		-> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
	"""

	:param epoch: a segment of an edf signal in the form of a 1d array
	:param fs: the frequency of the epoch in the form of an integer
	:param filter_order: ???
	:return: ???
	"""
	# Acknowledgement: Katrín Hera, M.Sc
	delta_band, theta_band, alpha_band, beta_band = eeg_freq_bands(epoch, fs, filter_order)

	window_size = int((fs / 2) - 1)
	delta_nonbiased = delta_band - signal.savgol_filter(delta_band, window_size, filter_order)
	theta_nonbiased = theta_band - signal.savgol_filter(theta_band, window_size, filter_order)
	alpha_nonbiased = alpha_band - signal.savgol_filter(alpha_band, window_size, filter_order)
	beta_nonbiased = beta_band - signal.savgol_filter(beta_band, window_size, filter_order)
	F_AF3, delta_PSD = signal.welch(delta_nonbiased, fs / 2, nperseg=len(delta_nonbiased))
	F_AF3, theta_PSD = signal.welch(theta_nonbiased, fs / 2, nperseg=len(theta_nonbiased))
	F_AF3, alpha_PSD = signal.welch(alpha_nonbiased, fs / 2, nperseg=len(alpha_nonbiased))
	F_AF3, beta_PSD = signal.welch(beta_nonbiased, fs / 2, nperseg=len(beta_nonbiased))

	return delta_PSD, theta_PSD, alpha_PSD, beta_PSD



import pyedflib as edf


def combine(file1, file2):
	# Take in both edf files
	# Open them both in PyEDFlib
	# Iterate through both files and find discrepancies
	# Create new file and add both to that one
	# If there already exists one signal with same name,
	# 	- create a new name for the signal to be able to compare the two
	# Check for min and max of the file
	# Return the new file

	if edf.highlevel.compare_edf(file1, file2):
		raise Exception("The contents of this file are the same")


	return None



import time
import pandas as pd
import pyedflib as edf
from typing import List


def get_signals(file, signalname: str or List[str] or List[dict]) -> List[int or float]:
	"""
	TODO: Allow the user to insert just the list of signals and the signalheaders and check for file type
	TODO: Document/Comment the code and make it more readable
	Args:
		file: The edf file that is to be imported
		signalname: The name of the signal that is wanted, can also be a list of signals that are wanted

	Returns: list of all the signals that were requested

	"""

	signal = []
	signals, signal_headers, header = edf.highlevel.read_edf(file)

	index = 0
	if signalname == 'all':
		for eeg_dict in signal_headers:
			signal.append(signals[index])
		index += 1

	# Check if signal name is either list or string
	if type(signalname) == list:

		# Check for duplicates in the list and remove them
		signalname = __check_duplicates_and_remove(signalname)
		for name in signalname:

			for eeg_dict in signal_headers:

				# If the
				if eeg_dict['label'] == name:
					signalname.remove(eeg_dict['label'])
					s = signals[index]
					signal.append(signals[index])

				if len(signalname) <= 0:
					break
				index += 1

	elif type(signalname) == str:

		for eeg_dict in signal_headers:

			if eeg_dict['label'] == signalname:
				s = signals[index]
				signal.append(signals[index])
			index += 1

	return signal


def __check_duplicates_and_remove(signal):
	"""
	Checks for duplicates and removes them

	Args:
		signal: A list of all signal labels

	Returns: list with all duplicates removed

	"""
	setOfElems = set()
	dupl = []

	for elem in signal:

		if elem in setOfElems:
			dupl.append(elem)

		else:
			setOfElems.add(elem)

	return list(setOfElems)



import pandas as pd
from numpy import ndarray

from . import bandpass


def eeg_freq_bands(eeg_signal: pd.DataFrame, sampling_freq: float or int,
                   filter_order: float or int) -> (ndarray, ndarray, ndarray, ndarray):
    """

    Args:
        eeg_signal: One type of signal from the edf file
        sampling_freq: The sampling frequency of the eeg signal
        filter_order:

    Returns: Four arrays that contain each bandpass of the most common frequencies

    """
    alpha_lower = 8
    alpha_upper = 14
    beta_lower = 14
    beta_upper = 50
    delta_lower = 0.5
    delta_upper = 4
    theta_lower = 4
    theta_upper = 8
    delta_band = bandpass(eeg_signal, delta_lower, delta_upper, sampling_freq, filter_order)
    theta_band = bandpass(eeg_signal, theta_lower, theta_upper, sampling_freq, filter_order)
    alpha_band = bandpass(eeg_signal, alpha_lower, alpha_upper, sampling_freq, filter_order)
    beta_band = bandpass(eeg_signal, beta_lower, beta_upper, sampling_freq, filter_order)
    return delta_band, theta_band, alpha_band, beta_band



import csv
from datetime import datetime, timedelta
from dateutil import parser
import numpy as np
import pandas as pd
from pytz import timezone


withings_signal_names = {
    'withings_sleep_state': 'raw_sleep-monitor_sleep-state',
    'withings_sleep_hr': 'raw_sleep-monitor_hr',
    'withings_watch_state': 'raw_tracker_sleep-state',
    'withings_watch_hr': 'raw_tracker_hr',
    'withings_sleep_movement_max': 'raw_sleep-monitor_maximum_movement',
    'withings_sleep_movement_pim': 'raw_sleep-monitor_movement_pim',
}


def get_e4_info(filename):
    with open(filename, 'r') as read_obj:
        csv_reader = csv.reader(read_obj)
        i = 0
        for row in csv_reader:
            if i == 0:
                n = len(row)
                start_time = ['']*n
                for j in range(len(start_time)):
                    start_time[j] = datetime.utcfromtimestamp(float(row[j])).isoformat(
                        sep=' ',
                        timespec='milliseconds'
                    )
            elif i == 1:
                n = len(row)
                freq = [0.0]*n
                for j in range(len(freq)):
                    freq[j] = float(row[j])
            else:
                break
            i += 1
        return start_time, freq


def e4_to_dataframe(signal_name, path):
    """

    Args:
        signal_name:
        path:

    Returns:

    """
    file = path + signal_name + '.csv'
    start, freq = get_e4_info(file)
    period = 1000000/(freq[0]) # In microseconds

    signal_df = pd.read_csv(file, skiprows=2, header=None)

    daterange = pd.date_range(
            start = start[0],
            periods = len(signal_df),
            freq = str(int(period))+'U'
            )

    df = pd.DataFrame(daterange, columns=['time'])
    df = pd.concat([df, signal_df], axis=1)
    df['datetime'] = pd.to_datetime(df['time'])
    df = df.set_index('datetime')
    df.drop(['time'], axis=1, inplace=True)

    return df


def withings_csv_to_dataframe(signal, path):

    signalname = withings_signal_names[signal]
    file = path + signalname + '.csv'
    with open(file, 'r') as read_obj:
        csv_reader = csv.reader(read_obj)
        i=0
        for row in csv_reader:
            row = np.array(row)
            if i==0:
                signal_np = np.array([row])
            else:
                starttime = parser.parse(row[0])
                vals = np.fromstring(row[2][1:-1], sep=',')
                durations = np.fromstring(row[1][1:-1], sep=',')
                for j in range(len(vals)):
                    val = vals[j]
                    duration = durations[j]
                    starttime = starttime+timedelta(seconds=int(duration))
                    starttime = starttime.astimezone(timezone('GMT')) # GMT for Iceland
                    signal_np = np.append(signal_np,[[starttime, duration, val]], axis=0)
            i = i+1

        df = pd.DataFrame(data=signal_np[1:,1::], index=signal_np[1:,0], columns=signal_np[0,1:]).sort_index()

        return df


from random import random
from typing import List
import numpy as np


def _segment_fs(signal: List[float or int], fs: int):
	# Maybe have this in a separate file for everyone to use
	return [signal[x:x + fs] for x in range(0, len(signal), fs)]


def lower_frequency(signal: List[float or int], method: str, new_fs: int) -> List[float or int]:
	"""
	Lowers the frequency of the given signal by some given method
	Args:
		signal: List of time series signals
		method: The method how the frequency should be lowered
		new_fs: The new frequency of the signal

	Returns: List of new values corresponding to the method given

	"""
	new_signal = []
	segmented_signal = _segment_fs(signal, new_fs)
	if method.lower() == 'min':
		for index in range(0, len(segmented_signal)):
			the_lowest = min(segmented_signal[index])
			new_signal.append(the_lowest)

	elif method.lower() == 'max':
		for index in range(0, len(segmented_signal)):
			the_lowest = max(segmented_signal[index])
			new_signal.append(the_lowest)

	elif method.lower() == 'mean':
		for index in range(0, len(segmented_signal)):
			the_lowest = np.mean(segmented_signal[index])
			new_signal.append(the_lowest)

	elif method.lower() == 'median':
		for index in range(0, len(segmented_signal)):
			the_lowest = np.median(segmented_signal[index])
			new_signal.append(the_lowest)

	return new_signal


if __name__ == "__main__":
	a = [round(random()*100, 1) for _ in range(1000)]
	b = lower_frequency(a, method='mean', new_fs=100)
	print(b)
	print(len(b))



from .math import *
from .get_segment import *
from .core import *
from .math import welch_psd
from .filter import *
from .get_edf import *
from .get_signals import *
from .lower_frequency import *
from .pair_signals import *
from .read_mat import *
from .wavelet import *


from typing import Union, List
import numpy as np

"""
TODO: Implement the functions and add them to the adaptive segmentations function
TODO: Document and comment the functions
TODO: Move find_transients function to some other file
"""

def _signal_afc(N: Union[int, float], i: Union[int, float],
				signals: Union[List[Union[int, float]], np.ndarray]):
	R_sum = 0

	for n in range(1, N - i):
		R_sum += signals[n] * signals[n + i]

	R_sum = R_sum * (1 / N)

	return R_sum


def _lp_filter_creation():
	return None


def _est_signal_value(signal, p, a, n):
	the_sum = 0
	for index in range(1, p):
		# TODO: check if the n is smaller than p and make some error handling for that, this should be able to go back in time
		the_sum += a[index] * signal[n - index]
	return -1 * the_sum

def error_of_signal():
	return


def _calculate_pe_val():
	return None


def _calculate_pe_acf():
	return None


def _calculate_sem():
	return None


def find_transients(hello):
	# This is part of the feature extraction process
	return None

def adaptive_segmentation(signals: Union[List[Union[int, float]], np.ndarray]):
	"""

	:param signals:
	:return:
	"""
	# Store segment length - Store both the onset and the duration
	# Store signal predictor
	# Store corrective predictor
	N = None or 10 # maybe 50?
	i = None or 666 or 56

	# Step 1:
	# 	a. Compute new signal ACF
	_signal_afc(N, i, signals)

	# 	b. adapt LP filter
	# Step 2:

	return None


from scipy import signal
import numpy as np
from typing import List

"""
TODO: Make the code more readable and document/comment it
"""


def low_pass_filter(data: List[int or float], cutoff: int or float,
					frequency: int or float, order=5) -> np.ndarray:
	"""
	:param data: Array of integers or float
	:param cutoff: The frequency that is to be filtered through --> integer or float
	:param frequency: The frequency of the data --> integer or float
	:param order: The order of the filter --> integer
	:return: An array of filtered data using the Buttersworth method --> Numpy ndarray
	"""
	b, a = butter_pass(cutoff, frequency, btype='lowpass', order=order)
	y = signal.lfilter(b, a, data)
	return y


def high_pass_filter(data: List[int or float] or np.ndarray, cutoff: int or float,
					 frequency: int or float, order=5) -> np.ndarray:
	"""
	:param data: Array of integers or float
	:param cutoff: The frequency that is to be filtered through --> integer or float
	:param frequency: The frequency of the data --> integer or float
	:param order: The order of the filter --> integer
	:return: An array of filtered data using the Buttersworth method --> Numpy ndarray
	"""
	b, a = butter_pass(cutoff, frequency, btype='highpass', order=order)
	y = signal.filtfilt(b, a, data)
	return y


def butter_pass(cutoff: int or float, frequency: int or float,
				btype: str, order: int = 5) -> (np.ndarray, np.ndarray):
	"""

	:param cutoff: The frequency that is to be filtered through --> integer or float
	:param frequency: The frequency of the data --> integer or float
	:param order: The order of the filter --> integer
	:param btype: The type of filtering --> string
	:return: The denominator and the numerator polynomials of the IIR filter --> Numpy ndarray, numpy ndarray
	"""
	nyq = 0.5 * frequency
	normal_cutoff = cutoff / nyq
	b, a = signal.butter(order, normal_cutoff, btype=btype, analog=False)
	return b, a



from typing import List

import pandas as pd


# TODO: change the actual code to my code
# TODO: comment
# TODO: pretty up the code through typing conventions
# TODO: UnitTests


def get_segment(index1: int, index2: int, signal: pd.DataFrame) -> List[int or float]:
	"""
	Args:
		index1: The start of the epoch
		index2: The end of the epoch
		signal: The data

	Returns: An array of segmented data
	"""
	signal_list = []

	for i in range(index2 - index1):
		# while index < len(signal)
		signal_list.append(signal[index1 + i])
	return signal_list



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Arora",
    version=os.getenv(key="PACKAGE_VERSION", default="0.0.5"),
    long_description=long_description,
    long_description_content="text/markdown",
    author="Sleep Revolution",
    author_email="sleeprevolution@ru.is",
    description="A tool for sleep researcher to preprocess, work, analyze and visualize data",
    packages=setuptools.find_packages(include=['Arora', 'Arora.*']),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8"
        ],
    install_requires=["numpy", "pandas", "pyEDFlib", "scipy"],
    python_requires=">=3.8"
)



"""
THIS IS A WIP file and is subject to change, will probably be turned into a directory later on
when more features have been implemented
"""
import datetime
from typing import List

import pandas as pd
from pandas import DataFrame


# TODO: comment
# TODO: pretty up the code through typing conventions
# TODO: UnitTests


def get_sample_segment(onset: int, signals: List[int or float], freq: float, duration=30, time_unit="sec"):
    """
    TODO: Allow the user to decide if he only wants to get the signals or also the Dataframe
    :param onset: start of recording in seconds (int)
    :param signals: list of EEG signals (float)
    :param freq: sample frequency of the EEG signals (int)
    :param duration: end of recording in seconds (int)
    :param time_unit: the unit of time for the duration (str)
    :return: a list of EEG signals and pandas DataFrame including the beginning and end index of the sample
    """
    if time_unit == "min":
        duration *= 60
    elif time_unit == "hours":
        duration *= 3600

    signal_list = []
    index_dataframe = pd.DataFrame()

    # find the beginning index by transferring time into index using the sampling frequency
    beg_index = int(onset * freq)

    # find the end index by figuring out the
    end_index = int((onset + duration) * freq)

    # end index might be out of scope of the list in which case we'll return as much as we can
    if end_index > len(signals):
        # TODO: warning/error handling over here
        print("Warning: Duration is beyond the scope of your list, data might not be reliable")

        signal_list = signals[beg_index:]
        end_index = len(signals)
    else:
        signal_list = signals[beg_index: end_index + 1]

    index_dataframe['beg_index'] = [beg_index]
    index_dataframe['end_index'] = [end_index]

    return signal_list, index_dataframe


def epochize(data: pd.DataFrame, channel_names: List[str], epoch_len: int,
             sampling_freq: int or float, start_timestamp: datetime) -> pd.DataFrame:
    """
    Args:
        data: EDF file turned into an array -> pd.DataFrame
        channel_names: The names of the channels to be used -> list[str]
        epoch_len: How long the epoch should be - will be changed to duration later on -> integer
        sampling_freq: The sampling frequency of the data -> float or integer
        start_timestamp: ??? - Will be removed later on

    Returns: A dataframe with the epochs - will be changed later on to a tuple with the epoch,
            start and duration
    """
    # WIP, will need to change the return, and some things in the function so that it will fit
    # previous work
    # Acknowledgement: Katrín Hera, M.Sc. student

    # Find the length of the signal
    l = len(data)
    # Find the number of data points
    number = int(epoch_len * sampling_freq)
    # initialize the epoch list
    epochs = []
    # add the epochs to the list
    for x in range(0, l, int(epoch_len * sampling_freq)):
        epochs.append(data[x:x + number])
    # initialize a list of timestamps at the beginning of each epoch
    timestamps_epoch_start = []
    for i in range(int(len(data) / (epoch_len * sampling_freq))):
        # make start timestamps for all the subsequent epochs and add to list
        next_timestamp = start_timestamp + datetime.timedelta(seconds=epoch_len * i)
        timestamps_epoch_start.append(next_timestamp)
    # make dataframe with timestamps and epoch
    df: DataFrame = pd.DataFrame()
    df['Epoch start'] = timestamps_epoch_start
    df[channel_names] = epochs  # temporarilly commented out for debugging

    # df['eh_test'] = pd.to_datetime(df['start times'])
    # df = df.set_index('Epoch start')
    return df
